
/*
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
 

#ifdef _WIN32
 #define _CRT_SECURE_NO_WARNINGS
#endif

#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <iostream>
#include <string>

#ifdef _WIN32
 #include "..\..\..\include\pkcs11\cryptoki_win32.h"
#else
 #include "..\..\..\include\pkcs11\cryptoki_linux.h"
#endif

// The default user pin of the token
#define TOKEN_USER_PIN "12345678"

// The ID for the keypair
#define KEY_ID 0xA1


void check_return_value(CK_RV rv, const char *message)
{
    if (rv != CKR_OK) 
    {
        fprintf(stderr, "Error at %s: %u\n", message, (unsigned int)rv);
        exit(EXIT_FAILURE);
    }
}

CK_SLOT_ID get_slot()
{
    CK_RV rv = CKR_OK;
    CK_SLOT_ID slotId = 0;
    CK_ULONG slotCount = 10;
    CK_SLOT_ID *slotIds = (CK_SLOT_ID *) calloc(slotCount, sizeof(CK_SLOT_ID));

    rv = C_GetSlotList(CK_TRUE, slotIds, &slotCount);
    check_return_value(rv, "get slot list");

    if (slotCount < 1) 
    {
        fprintf(stderr, "Error; could not find any slots\n");
        exit(1);
    }

    slotId = slotIds[0];
    free(slotIds);
    printf("slot count: %d\n", (int)slotCount);
    return slotId;
}

CK_SESSION_HANDLE start_session(CK_SLOT_ID slotId)
{
    CK_RV rv = CKR_OK;
    CK_SESSION_HANDLE session;
    rv = C_OpenSession(slotId, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session);
    check_return_value(rv, "open session");
    return session;
}

void login(CK_SESSION_HANDLE session, CK_BYTE *pin)
{
    CK_RV rv = CKR_OK;
    if (pin) 
    {
        rv = C_Login(session, CKU_USER, pin, strlen((char *)pin));
        check_return_value(rv, "log in");
    }
}

void create_key_pair(CK_SESSION_HANDLE session)
{
    CK_RV rv;
    CK_OBJECT_HANDLE publicKey, privateKey;
    CK_MECHANISM mechanism = {
        CKM_RSA_PKCS_KEY_PAIR_GEN, NULL_PTR, 0
    };
    CK_ULONG modulusBits = 2048;
    CK_BYTE publicExponent[] = { 1, 0, 1 };    // The public exponent 0x010001 (65537 in decimal)
    CK_BYTE subject[] = "testkey";
    CK_BYTE id[] = {KEY_ID};
    CK_BBOOL bTrue = CK_TRUE;
    CK_ATTRIBUTE publicKeyTemplate[] = {
        {CKA_ID, id, sizeof(id)},
        {CKA_LABEL, subject, 7},
        {CKA_TOKEN, &bTrue, sizeof(bTrue)},
        {CKA_ENCRYPT, &bTrue, sizeof(bTrue)},
        {CKA_VERIFY, &bTrue, sizeof(bTrue)},
        {CKA_WRAP, &bTrue, sizeof(bTrue)},
        {CKA_MODULUS_BITS, &modulusBits, sizeof(modulusBits)},
        {CKA_PUBLIC_EXPONENT, publicExponent, 3}
    };
    CK_ATTRIBUTE privateKeyTemplate[] = {
        {CKA_ID, id, sizeof(id)},
        {CKA_LABEL, subject, 7},
        {CKA_TOKEN, &bTrue, sizeof(bTrue)},
        {CKA_PRIVATE, &bTrue, sizeof(bTrue)},
        {CKA_SENSITIVE, &bTrue, sizeof(bTrue)},
        {CKA_DECRYPT, &bTrue, sizeof(bTrue)},
        {CKA_SIGN, &bTrue, sizeof(bTrue)},
        {CKA_UNWRAP, &bTrue, sizeof(bTrue)}
    };

    rv = C_GenerateKeyPair(
        session,
        &mechanism,
        publicKeyTemplate, 8,
        privateKeyTemplate, 8,
        &publicKey,
        &privateKey
    );
    check_return_value(rv, "generate key pair");

    return;
}

void digest_data(CK_SESSION_HANDLE session, FILE *data_file, CK_BYTE **digestOut, CK_ULONG *digestLenOut)
{
    CK_MECHANISM digest_mechanism;
    CK_RV rv;

    size_t dataLen = 0;
    size_t dataTotal = 0;
    CK_BYTE data[1024];

    if (!data_file) 
    {
        fprintf(stderr, "Error, no file handle in digest_data\n");
        exit(3);
    }

    CK_ULONG digestLen = 20;    // 20 bytes required for SHA-1 hash
    CK_BYTE *digest = (CK_BYTE *)malloc(digestLen);

    digest_mechanism.mechanism = CKM_SHA_1;
    digest_mechanism.pParameter = NULL;
    digest_mechanism.ulParameterLen = 0;

    rv = C_DigestInit(session, &digest_mechanism);
    check_return_value(rv, "digest init");

    fseek (data_file, 0 , SEEK_END);
    long size = ftell(data_file);
    printf("File is %ld bytes\n", size);
    rewind(data_file);
    dataLen = fread(data, 1, sizeof(data), data_file);
    while (dataLen > 0) 
    {        
        rv = C_DigestUpdate(session, data, dataLen);
        check_return_value(rv, "digest update");
        dataTotal += dataLen;
        dataLen = fread(data, 1, sizeof(data), data_file);
    }
    printf("Added %ld bytes to digest\n", dataTotal);

    rv = C_DigestFinal(session, digest, &digestLen);
    check_return_value(rv, "digest final");

    *digestOut = digest;
    *digestLenOut = digestLen;

    return;
}

CK_OBJECT_HANDLE get_key(CK_SESSION_HANDLE session, CK_OBJECT_CLASS keyClass, CK_BYTE *id, size_t id_len)
{
    CK_RV rv;
    CK_ATTRIBUTE t[] = {
        { CKA_CLASS, &keyClass, sizeof(keyClass) },
        { CKA_ID, id, id_len }
    };
    CK_ULONG objectCount;
    CK_OBJECT_HANDLE object;

    rv = C_FindObjectsInit(session, t, 1);
    check_return_value(rv, "Find objects init");

    rv = C_FindObjects(session, &object, 1, &objectCount);
    check_return_value(rv, "Find first object");

    if (objectCount > 0) 
    {
        rv = C_FindObjectsFinal(session);
        check_return_value(rv, "Find objects final");          
    } 
    else 
    {
        fprintf(stderr, "Private key not found\n");
        exit(2);
    }

    return object;
}

void sign_data_example(CK_SESSION_HANDLE session)
{
    // First we ask what file they want to calculate the digital signature for
    FILE *data_file = NULL;
    std::string filename;
    do
    {
        std::cout << "File name to calculate digital signature of:" << std::endl;        
        std::getline(std::cin, filename);
        data_file = fopen(filename.c_str(), "rb");
    } while (!data_file);

    printf("Successfully opened \"%s\"\n", filename.c_str());

    
    // Now we calculate the digital signature of the data in the file.
    // The steps are:
    // 1. Calculate the digest (hash) of the data
    // 2. Sign the digest using the private key


    CK_RV rv;
    CK_BYTE id[] = { KEY_ID };
    CK_OBJECT_HANDLE privateKey = get_key(session, CKO_PRIVATE_KEY, id, 1);        // Get the private key (Note: pass CKO_PUBLIC_KEY to obtain the public key)
    CK_MECHANISM sign_mechanism;

    CK_ULONG signatureLen = 512;
    CK_BYTE *signature = (CK_BYTE *)malloc(signatureLen);

    sign_mechanism.mechanism = CKM_RSA_PKCS;
    sign_mechanism.pParameter = NULL;
    sign_mechanism.ulParameterLen = 0;

    CK_ULONG digestLen = 0;
    CK_BYTE *digest = NULL;
    digest_data(session, data_file, &digest, &digestLen);

    rv = C_SignInit(session, &sign_mechanism, privateKey);
    check_return_value(rv, "sign init new");

    rv = C_Sign(session, digest, digestLen, signature, &signatureLen);
    check_return_value(rv, "sign final");

    if (signatureLen > 0) 
    {
        printf("Signature is:");
        for (CK_ULONG i=0; i<signatureLen; i++)
            printf("%02x ", signature[i]);
        printf("\n");
    }

    free(digest);
    free(signature);
     
    return;
}

bool key_exists(CK_SESSION_HANDLE session, CK_OBJECT_CLASS keyClass, CK_BYTE *id, size_t id_len)
{
    CK_RV rv;
    CK_BBOOL bTrue = CK_TRUE;
    CK_ATTRIBUTE t[] = {
        { CKA_ID, id, id_len },
        { CKA_CLASS, &keyClass, sizeof(keyClass) },
        { CKA_TOKEN, &bTrue, sizeof(bTrue) },
        { CKA_PRIVATE, &bTrue, sizeof(bTrue) },
    };
    CK_ULONG objectCount;
    CK_OBJECT_HANDLE object;

    rv = C_FindObjectsInit(session, t, 1);
    check_return_value(rv, "Find objects init");

    rv = C_FindObjects(session, &object, 1, &objectCount);
    check_return_value(rv, "Find first object");

    if (objectCount > 0) 
    {
        return true;
    } 
    
    return false;
}

int main(int argc, char **argv)
{
    CK_RV rv = CKR_OK;
    CK_SLOT_ID slotID = 0;
    CK_SESSION_HANDLE hSession = 0;

    // Initialize the PKCS11 library
    rv = C_Initialize(NULL);
    check_return_value(rv, "pkcs11 initalize");

    // Get a slot containing a token
    slotID = get_slot();
    
    // Start a PKCS11 session with the token
    hSession = start_session(slotID);

    // Login to the token using the default pin defined at the top of this file
    login(hSession, (CK_BYTE*)TOKEN_USER_PIN);

    /*************************************************************/
    // Initialization is complete.
    // We can do something interesting using the token now!
    /*************************************************************/

    // Create a new key pair (unless it already exists)
    CK_BYTE key_id[] = { KEY_ID };
    if (key_exists(hSession, CKO_PUBLIC_KEY, key_id, 1) == false)
    {
        printf("Creating new key pair\n");
        // Create a new key pair
        create_key_pair(hSession);
        printf("Key pair created\n");
    }
    else
        printf("Key pair already exists\n");

    // Sign some data
    sign_data_example(hSession);

    /*************************************************************/

    // Cleanup
    rv = C_CloseSession(hSession);
    check_return_value(rv, "close session");

    C_Finalize(NULL);

    printf("Exiting successfully");
    return 0;
}



